// ========			messiahMayaManager.cpp			==============
//
// Copyright (C) 2002 pmG Worldwide LLC. All Rights Reserved.
//
// ===============================================================
//
// Description: 
//
// Implementation of messiahMayaManager.
//
// ===============================================================
//
// Version History:
//
// Written By: Ron Griswold
// Date: 4/18/2002
//
// Revised By:
// Date: 
//
// ===============================================================
#include "messiahMayaManager.h"

#include "Params.h"

#include "messiahCommonNodeOps.h"
#include "messiahDeformerNode.h"
#include "messiahTransformerNode.h"

#include "MH_System.h"
#include "MH_NodeLogFuncs.h"

#include <direct.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <maya/MPxNode.h>
#include <maya/MFnDependencyNode.h>
#include <maya/MPlug.h>
#include <maya/MFnStringData.h>
#include <maya/MFnNumericData.h>
#include <maya/MObject.h>
#include <maya/MFn.h>
#include <maya/MTime.h>
#include <maya/MTypeId.h>
#include <maya/MGlobal.h>

#define MESSIAH_HOST_MODULE
#include "messiahHOST.h"			// for scene buffer ops


extern char _messiahDir[_MAX_PATH];


// ===============================================================
// Initialize static variables
// ===============================================================
messiahMayaManager * messiahMayaManager::_inst = 0;

int	const messiahMayaManager::SceneLoading				= 0x1;
int	const messiahMayaManager::SceneSaving				= 0x2;
int	const messiahMayaManager::SceneClearing				= 0x4;
int	const messiahMayaManager::SceneActiveMessiahSession = 0x8;

// ===============================================================
// Get Instance
// ===============================================================
messiahMayaManager * messiahMayaManager::Inst()
{
	if( !_inst )
	{
		_inst = new messiahMayaManager;
	}

	return _inst;
}

// ===============================================================
// Scene State
// ===============================================================
int		messiahMayaManager::SceneState()
{
	return this->_sceneState;
}

int		messiahMayaManager::SceneState(int flag, bool state)
{
	if( state )
		this->_sceneState |= flag;
	else
		this->_sceneState ^= (this->_sceneState & flag ? flag : 0);

	return this->_sceneState;
}

// ===============================================================
// Add Node
// ===============================================================
void	messiahMayaManager::AddNode(MPxNode *node)
{
	this->_nodes.push_back(node);

	if( !(SceneState() & SceneActiveMessiahSession) )
	{
		// This is the first node to be added to the scene
		// this signals the start of a new messiah session
		// if the scene is not being loaded from file
		// we need to create a script node that will hold
		// the fxs buffer
		if( !(SceneState() & SceneLoading) )
		{
			MFnDependencyNode dNode;
			MStatus		stat;
			
			dNode.create(MM_MAN_ID, &stat);
			if(!stat)
				return;
			this->_sceneNode = dNode.userNode();
			dNode.setName(MString(MM_MAN_NAME));
			if(!stat)
				return;
		}
		
		SceneState(SceneActiveMessiahSession, true);
	}
}

// ===============================================================
// Remove Node
// ===============================================================
void	messiahMayaManager::RemoveNode(MPxNode *node)
{
	std::vector<MPxNode *>::iterator	it;

	it = this->_nodes.begin();

	while( it != this->_nodes.end() )
	{
		if( *it == node )
		{
			this->_nodes.erase(it);
			return;
		}

		it++;
	}

	if( this->_nodes.empty() )
	{
		SceneState(SceneActiveMessiahSession, false);
	}
}

void	messiahMayaManager::SceneNode(MPxNode *node)
{
	if(!node)
		return;

	this->_sceneNode = node;
}

// ===============================================================
// Load Callback
// ===============================================================
// Pre
void	messiahMayaManager::OnLoadPre		(void *data)
{
	// Set SceneState flag to indicate scene is loading
	MAN::Inst()->SceneState( messiahMayaManager::SceneLoading, true );
}
//Post
void	messiahMayaManager::OnLoadPost		(void *data)
{
	// Set SceneState flag to indicate scene has finished loading
	MAN::Inst()->SceneState( messiahMayaManager::SceneLoading, false );
	MStatus stat;
	
	// Load scene node ///////////////////////////////////////////////////
	if( MAN::Inst()->_sceneNode )
	if( !MAN::Inst()->_sceneNode->thisMObject().isNull() )
	{
		MFnDependencyNode dNode( MAN::Inst()->_sceneNode->thisMObject() );

		// fxs buffer
		MObject attr = dNode.attribute(MString(MM_ATR_FXSBUFF_S));
		MPlug	plug( MAN::Inst()->_sceneNode->thisMObject(), attr );

		if( plug.isNull() )
		{
			MGlobal::displayError(MString("Error reading messiah data from scene file"));
			return;
		}

		MString mfxs;
		char *fxs = 0;

		stat = plug.getValue(mfxs);
		if( !stat )
			MGlobal::displayError(MString("Failed to retrieve messiah scene buffer"));
		fxs = const_cast<char *>(mfxs.asChar());

		messiahSceneFileLoadBuff(fxs);

		// scale factor
		MHS::scaleFactor = 1.0;

		attr = dNode.attribute(MString(MM_ATR_SCALE_S));
		MPlug plug2( MAN::Inst()->_sceneNode->thisMObject(), attr );

		if( plug2.isNull() )
		{
			MGlobal::displayError(MString("Error reading scale factor from file"));
			return;
		}
		
		MString res;
		stat = plug2.getValue(res);
		MHS::scaleFactor = res.asDouble();

		messiahSceneLoad(MESSIAH_SCENE_LOAD);
	}////////////////////////////////////////////////////////////////////////////////////////

}

// ===============================================================
// Save Callback
// ===============================================================
// Pre
void	messiahMayaManager::OnSavePre		(void *data)
{
	if( MH_NodeLogFuncs::NumNodes() == 0 ) return;

	// Set SceneState flag to indicate scene is saving
	MAN::Inst()->SceneState( messiahMayaManager::SceneSaving, true );

	MStatus stat;

	// cycle through the nodes and set their load id's to match the id from the nodeLog ///////////////

	std::vector<MPxNode *>::iterator	it;
	for(it = MAN::Inst()->_nodes.begin(); it != MAN::Inst()->_nodes.end(); it++)
	{
		MTypeId mid = (*it)->typeId();
		unsigned int id = mid.id();
		
		messiahDeformer *		dnode = NULL;
		messiahTransformer *	xnode = NULL;
		commonNode *			cnode = NULL;
		
		// Cast the MPxNode * to a commonNode *
		// since commonNode is not derived from MPxNode
		// we must first go through the deformer and xformer node types
		if( id == MM_DISP_ID )
		{
			dnode = (messiahDeformer *)*it;
			cnode = (commonNode *)dnode;
		}
		else if( id == MM_MOVE_ID )
		{
			xnode = (messiahTransformer *)*it;
			cnode = (commonNode *)xnode;
		}
		
		if( cnode->_nodeLog )
		{
			int id = cnode->_nodeLog->ID;

			MPlug				loadId;
			MFnDependencyNode	dNode( (*it)->thisMObject() );

			loadId = dNode.findPlug( MM_ATR_LOAD_ID_L, &stat);		// Get loadId plug
			if(!stat)
			{
				MGlobal::displayError("failed to select loadId plug");
				return;
			}

			loadId.setValue( id );									// Set it's value to match that of the node log
		}// if nodeLog

	}////////////////////////////////////////////////////////////////////////////////////////////////////

	// Set the scene node (messiahNode) ///////////////////////////////////////////////
	MFnDependencyNode dNode( MAN::Inst()->_sceneNode->thisMObject() );

	// fxs buffer
	MObject attr = dNode.attribute(MString(MM_ATR_FXSBUFF_S));
	MPlug	plug( MAN::Inst()->_sceneNode->thisMObject(), attr );

	if( plug.isNull() )
	{
		MGlobal::displayError(MString("Error writing messiah data to scene file"));
		return;
	}

	char *fxs = 0;

	if( !(fxs = messiahSceneFileCreateBuff()) )
	{
		MGlobal::displayError(MString("Failed to retrieve scene information from messiah"));
		return;
	}

	stat = plug.setValue( fxs );

	if( !stat )
		MGlobal::displayError(MString("Failed to store messiah scene buffer"));

	messiahSceneFileClearBuff( fxs );

	// scale factor
	attr = dNode.attribute(MString(MM_ATR_SCALE_S));
	MPlug plug2( MAN::Inst()->_sceneNode->thisMObject(), attr );

	if( plug2.isNull() )
	{
		MGlobal::displayError(MString("Error writing messiah data to scene file"));
		return;
	}
	
	char tmpStr[20];
	sprintf(tmpStr, "%f", MHS::scaleFactor);
	plug2.setValue( tmpStr );

	//////////////////////////////////////////////////////////////////////////////////////////////////////

}
//Post
void	messiahMayaManager::OnSavePost		(void *data)
{

	// Set SceneState flag to indicate scene has finished saving
	MAN::Inst()->SceneState( messiahMayaManager::SceneSaving, false );

}

// ===============================================================
// Clear (New) Callback
// ===============================================================
// Pre
void	messiahMayaManager::OnClearPre		(void *data)
{
	MAN::Inst()->SceneState( messiahMayaManager::SceneClearing, true );

	// clear node_log
	NLF::ClearAllElements(2);
	
	// TODO: node's destructors should remove themselves from this list
	// therefore this call should be unnescessary.
	// Perhaps call it in OnClearPost as a precaution only.
	MAN::Inst()->_nodes.clear();
}
// Post
void	messiahMayaManager::OnClearPost		(void *data)
{
	MAN::Inst()->SceneState( messiahMayaManager::SceneClearing, false );

}

// ===============================================================
// Time Changed Callbacks
// ===============================================================
void	messiahMayaManager::OnTimeChange		(MTime &time, void *data)
{
	MAN::Inst()->_time = time.value();

	if( MH_System::Inst() )
	{
		// TODO: Add support for other frame rates
		MHS::Inst()->CheckTime( MAN::Inst()->CurrentTime() , 1);
	}
}
